import { FormulaError } from "../evaluator";
import { solveTable } from "../solver";
import { Table } from "../../lib/table";
import { BaseFunction } from "./__base";
import { ensureBoolean, ensureNumber, stripTable } from "./__utils";
export class VlookupFunction extends BaseFunction {
    constructor() {
        super(...arguments);
        this.example = "VLOOKUP(10003, A2:B26, 2, FALSE)";
        this.helpText = [
            "Searches vertically for the specified key in the first column of the range and returns the value of the specified cell in the same row.",
        ];
        this.helpArgs = [
            { name: "key", description: "Search key." },
            {
                name: "range",
                description: "A range for search",
            },
            {
                name: "index",
                description: "The index of the column in the range.",
            },
            {
                name: "is_sorted",
                description: "FALSE: Exact match. This is recommended. TRUE: Approximate match. Before you use an approximate match, sort your search key in ascending order. Otherwise, you may likely get a wrong return value.",
                option: true,
            },
        ];
    }
    validate() {
        if (this.bareArgs.length !== 3 && this.bareArgs.length !== 4) {
            throw new FormulaError("#N/A", "Number of arguments for VLOOKUP is incorrect.");
        }
        if (this.bareArgs[0] instanceof Table) {
            this.bareArgs[0] = stripTable(this.bareArgs[0]);
        }
        if (!(this.bareArgs[1] instanceof Table)) {
            throw new FormulaError("#REF!", "2nd argument must be range");
        }
        this.bareArgs[2] = ensureNumber(this.bareArgs[2]);
        this.bareArgs[3] = ensureBoolean(this.bareArgs[3], true);
    }
    main(key, range, index, isSorted) {
        var _a, _b, _c, _d;
        const matrix = solveTable({ table: range });
        if (isSorted) {
            let last = -1;
            for (let y = 0; y <= range.getNumRows(); y++) {
                const v = (_a = matrix[y]) === null || _a === void 0 ? void 0 : _a[0];
                if (v == null) {
                    continue;
                }
                if (v <= key) {
                    last = y;
                }
                else {
                    break;
                }
            }
            if (last !== -1) {
                return (_b = matrix[last]) === null || _b === void 0 ? void 0 : _b[index - 1];
            }
        }
        else {
            for (let y = 0; y <= range.getNumRows(); y++) {
                if (((_c = matrix[y]) === null || _c === void 0 ? void 0 : _c[0]) === key) {
                    return (_d = matrix[y]) === null || _d === void 0 ? void 0 : _d[index - 1];
                }
            }
        }
        throw new FormulaError("#N/A", `No values found for '${key}'.`);
    }
}
//# sourceMappingURL=vlookup.js.map