import { parseFromTimeZone } from "date-fns-timezone";
import { TimeDelta } from "../lib/time";
const BOOLS = { true: true, false: false };
export class Parser {
    constructor(props) {
        this.parseFunctions = [
            this.number,
            this.timedelta,
            this.date,
            this.bool,
        ];
        this.applyMixins(props === null || props === void 0 ? void 0 : props.mixins);
        if (props == null) {
            return;
        }
        const { condition, complement } = props;
        this.condition = condition;
        this.complement = complement;
    }
    applyMixins(mixins) {
        if (mixins == null) {
            return;
        }
        for (const mixin of mixins) {
            for (const key in mixin) {
                // @ts-ignore
                this[key] = mixin[key];
            }
        }
    }
    callback(parsed, cell) {
        return parsed;
    }
    parse(value, cell) {
        try {
            const parsed = this._parse(value, cell);
            return Object.assign(Object.assign({}, cell), { value: parsed });
        }
        catch (e) {
            return Object.assign(Object.assign({}, cell), { value: e });
        }
    }
    _parse(value, cell) {
        if (this.condition && !this.condition(value)) {
            const result = this.complement ? this.complement(value) : value;
            return this.callback(result, cell);
        }
        if (value[0] === "'") {
            return this.callback(value, cell);
        }
        for (let i = 0; i < this.parseFunctions.length; i++) {
            const result = this.parseFunctions[i](value, cell);
            if (result != null) {
                return this.callback(result, cell);
            }
        }
        if (value === "") {
            return this.callback(null, cell);
        }
        return this.callback(value, cell);
    }
    bool(value, cell) {
        return BOOLS[value.toLowerCase()];
    }
    number(value, cell) {
        const m = value.match(/^-?[\d.]+$/);
        if (m != null &&
            value.match(/\.$/) == null &&
            (value.match(/\./g) || []).length <= 1) {
            return parseFloat(value);
        }
    }
    timedelta(value, cell) {
        if (value.length < 4 || isNaN(value[value.length - 1])) {
            return;
        }
        {
            const match = value.match(/^([+-]?)(\d+):(\d{2})$/);
            if (match) {
                const [, _sign, hours, minutes] = match;
                const sign = _sign === "-" ? -1 : 1;
                return TimeDelta.create(sign * Number(hours), sign * Number(minutes));
            }
        }
        {
            const match = value.match(/^([+-]?)(\d+):(\d{2}):(\d{2})$/);
            if (match) {
                const [, _sign, hours, minutes, seconds] = match;
                const sign = _sign === "-" ? -1 : 1;
                return TimeDelta.create(sign * Number(hours), sign * Number(minutes), sign * Number(seconds));
            }
        }
        {
            const match = value.match(/^([+-]?)(\d+):(\d{2}):(\d{2})\.(\d+)$/);
            if (match) {
                const [, _sign, hours, minutes, seconds, msecs] = match;
                const sign = _sign === "-" ? -1 : 1;
                return TimeDelta.create(sign * Number(hours), sign * Number(minutes), sign * Number(seconds), sign * Number(msecs));
            }
        }
    }
    date(value, cell) {
        const first = value[0];
        if (first == null || first.match(/[JFMASOND0-9]/) == null) {
            return;
        }
        if (value[value.length - 1].match(/[0-9Z]/) == null) {
            return;
        }
        let timeZone = "UTC";
        try {
            timeZone = Intl.DateTimeFormat().resolvedOptions().timeZone;
        }
        catch (e) { }
        const d = parseFromTimeZone(value, { timeZone });
        if (d.toString() === "Invalid Date") {
            return;
        }
        return d;
    }
}
export const defaultParser = new Parser();
//# sourceMappingURL=core.js.map