import { Table } from "../lib/table";
import { solveFormula } from "../formula/solver";
import { FormulaError } from "../formula/evaluator";
import { p2a } from "../lib/converters";
import { TimeDelta } from "../lib/time";
import { format as formatDate } from "date-fns";
export class Renderer {
    constructor(props) {
        this.datetimeFormat = "yyyy-MM-dd HH:mm:ss";
        this.dateFormat = "yyyy-MM-dd";
        this.timeDeltaFormat = "HH:mm";
        this.applyMixins(props === null || props === void 0 ? void 0 : props.mixins);
        if (props == null) {
            return;
        }
        const { condition, complement } = props;
        this.condition = condition;
        this.complement = complement;
    }
    applyMixins(mixins) {
        if (mixins == null) {
            return;
        }
        for (const mixin of mixins) {
            for (const key in mixin) {
                // @ts-ignore
                this[key] = mixin[key];
            }
        }
    }
    render(table, point, writer) {
        var _a;
        const address = p2a(point);
        const cache = table.getSolvedCache(address);
        const value = cache || ((_a = table.getByPoint(point)) === null || _a === void 0 ? void 0 : _a.value);
        const { y, x } = point;
        return this._render(value, table.trim({ top: y, left: x, bottom: y, right: x }), writer);
    }
    _render(value, table, writer) {
        var _a;
        if (this.condition && !this.condition(value)) {
            return this.complement ? this.complement(value) : this.stringify(value);
        }
        switch (typeof value) {
            case "object":
                if (value instanceof Date) {
                    return this.date(value, writer);
                }
                if (value instanceof TimeDelta) {
                    return this.timedelta(value, writer);
                }
                if (value == null) {
                    return this.null(value, writer);
                }
                if (value instanceof Table) {
                    return this._render((_a = value.getByPoint({ y: value.top, x: value.left })) === null || _a === void 0 ? void 0 : _a.value, table, writer);
                }
                if (Array.isArray(value)) {
                    return this.array(value, writer);
                }
                if (value instanceof FormulaError) {
                    throw value;
                }
                return this.object(value, writer);
            case "string":
                return this.string(value, table, writer);
            case "number":
                return this.number(value, writer);
            case "boolean":
                return this.bool(value, writer);
            case "undefined":
                return this.undefined(value, writer);
            case "function":
                return value();
        }
        return "";
    }
    stringify(cell) {
        const { value } = cell;
        if (value instanceof Date) {
            return this.date(value);
        }
        if (value instanceof TimeDelta) {
            return this.timedelta(value);
        }
        if (value == null) {
            return "";
        }
        return value.toString();
    }
    string(value, table, writer) {
        if (value[0] === "'") {
            return value.substring(1);
        }
        if (value[0] === "=") {
            const result = solveFormula({ value, table, raise: true });
            if (result == null) {
                // @ts-ignore
                return this[String(result)](result);
            }
            if (result.constructor.name === "Boolean") {
                return String(result).toUpperCase();
            }
            if (result.constructor.name === "Date") {
                return this.date(result);
            }
            return this._render(result, table, writer);
        }
        return value;
    }
    bool(value, writer) {
        return value ? "TRUE" : "FALSE";
    }
    number(value, writer) {
        if (isNaN(value)) {
            return "NaN";
        }
        return value;
    }
    date(value, writer) {
        if (value.getHours() + value.getMinutes() + value.getSeconds() === 0) {
            return formatDate(value, this.dateFormat);
        }
        return formatDate(value, this.datetimeFormat);
    }
    timedelta(value, writer) {
        return value.stringify(this.timeDeltaFormat);
    }
    array(value, writer) {
        return value.map((v) => this.stringify({ value: v })).join(",");
    }
    object(value, writer) {
        return JSON.stringify(value);
    }
    null(value, writer) {
        return "";
    }
    undefined(value, writer) {
        return "";
    }
}
export const defaultRenderer = new Renderer();
//# sourceMappingURL=core.js.map